import functools
import inspect
import jaeger_client
import atexit
import opentracing
from opentracing.scope_managers import contextvars


def close_tracer():
    opentracing.tracer.close()


def init_tracer(config: dict):
    config = jaeger_client.Config(
        config=config,
        validate=True,
        scope_manager=contextvars.ContextVarsScopeManager(),
        service_name=config["service_name"],
    )
    global_tracer = config.new_tracer()
    opentracing.set_global_tracer(global_tracer)
    atexit.register(close_tracer)


def trace(span_name: str, **tags):
    def inner(func):
        def _normal_func():
            return opentracing.tracer.start_active_span(
                span_name,
                child_of=opentracing.tracer.active_span,
            )

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            with _normal_func() as scope:
                for tag_key, tag_value in tags.items():
                    scope.span.set_tag(tag_key, tag_value)
                return await func(*args, **kwargs)

        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            with _normal_func() as scope:
                for tag_key, tag_value in tags.items():
                    scope.span.set_tag(tag_key, tag_value)
                return func(*args, **kwargs)

        if not inspect.iscoroutinefunction(func):
            return sync_wrapper
        return async_wrapper
    return inner



import abc
import peewee
import contextlib
import atexit
import flask
import opentracing

try:
    from dbutils import steady_db
except ImportError:
    from DBUtils import SteadyDB as steady_db


from ..utils import proxy
from ..backend import (
    context,
    mysql,
    cache,
)
from ..settings import ConfigPartition
from ..sample import jaeger

__all__ = ["ConfigManager", "BaseModel", "init_jaeger", "init_app"]


class ConfigManager(abc.ABC):
    @abc.abstractmethod
    def get_config(self, key: ConfigPartition) -> dict:
        raise NotImplementedError


class PoolMysqlDatabase(peewee.MySQLDatabase):

    def __init__(self, connect: steady_db.SteadyDBConnection, *args, **kwargs):
        self.connect_ = connect
        super(PoolMysqlDatabase, self).__init__(*args, **kwargs)

    def sequence_exists(self, seq):
        pass

    def _connect(self):
        return self.connect_


class BaseModel(peewee.Model):
    class Meta:
        database = proxy.LocalProxy(flask.g, "database")


tenant_envs = context.TenantEnvironments(max_size=10)


@atexit.register
def shut_down_db_pool():
    tenant_envs.close()


@contextlib.contextmanager
def build_context(token: context.TenantToken):
    tenant_item = tenant_envs[token]
    try:
        flask.g.database = PoolMysqlDatabase(
            connect=tenant_item.databasemetadata.connection(),
            database=token,
        )
        flask.g.redis = tenant_item.cachemetadata
        yield None
    except BaseException as e:
        raise e
    finally:
        flask.g.database.close()
        flask.g.redis.close()


def init_jaeger(
    app: flask.Flask,
    config_manager: ConfigManager,
):
    jaeger.init_tracer(config_manager.get_config(ConfigPartition.jaeger))

    def before_request():
        request = flask.request
        span_context = opentracing.tracer.extract(
            format=opentracing.Format.HTTP_HEADERS,
            carrier=dict(request.headers),
        )
        flask.g.jaeger_scope = opentracing.tracer.start_active_span(
            request.url.path,
            child_of=span_context)

    def after_request(response: flask.Response):
        headers_data = dict(response.headers)
        opentracing.tracer.inject(
            span_context=flask.g.jaeger_scope.span.context,
            format=opentracing.Format.HTTP_HEADERS,
            carrier=headers_data,
        )
        for key, value in headers_data.items():
            response.headers.add(_key=key, _value=value)
        flask.g.jaeger_scope.close()
        del flask.g.jaeger_scope

    app.before_request(before_request)
    app.after_request(after_request)


def init_app(
    app: flask.Flask,
    config_manager: ConfigManager,
):

    def before_request():
        mysql_config = config_manager.get_config(ConfigPartition.mysql)
        redis_config = config_manager.get_config(ConfigPartition.redis)
        token = context.TenantToken()
        token.organize(mysql.DataBaseMetaDataBuilder(**mysql_config))
        token.organize(cache.CacheMetaDataBuilder(**redis_config))
        tenant_item = tenant_envs[token]
        flask.g.database = PoolMysqlDatabase(
            connect=tenant_item.databasemetadata.connection(),
            database=token,
        )
        flask.g.redis = tenant_item.cachemetadata

    def after_request(response: flask.Response):
        flask.g.database.close()
        del flask.g.database
        del flask.g.redis
        return response

    app.before_request(before_request)
    app.after_request(after_request)



import abc
from typing import Dict, List, Generic, TypeVar

from ..utils import (
    lru,
)


class MetaData(abc.ABC):

    @abc.abstractmethod
    def discard(self):
        raise NotImplementedError


metadata = TypeVar("metadata", bound=MetaData)


class MetaDataBuilder(abc.ABC, Generic[metadata]):

    def build(self) -> metadata:
        raise NotImplementedError

    def __hash__(self):
        return self.hash()

    def __eq__(self, other: "MetaDataBuilder"):
        return self.hash() == other.hash()

    @abc.abstractmethod
    def hash(self) -> str:
        raise NotImplementedError


class TenantContext(lru.Element):

    __slots__ = ("__meta_map",)

    def __init__(self):
        self.__meta_map: Dict[str, MetaData] = {}

    def inject(self, meta_data: MetaData):
        self.__meta_map.update({
            type(meta_data).__name__.lower(): meta_data
        })

    def __getattr__(self, item):
        return self.__meta_map[item]

    def discard(self):
        for meta_data in self.__meta_map.values():
            meta_data.discard()


class TenantToken(lru.Token):

    __slots__ = ("__builders",)

    def __init__(self):
        self.__builders: List[MetaDataBuilder] = []

    def organize(self, meta_builder: MetaDataBuilder):
        self.__builders.append(meta_builder)

    def hash(self) -> int:
        return tuple([
            _.hash() for _ in self.__builders
        ]).__hash__()

    def build_context(self) -> TenantContext:
        tenant_context = TenantContext()
        for builder in self.__builders:
            tenant_context.inject(builder.build())
        return tenant_context


class TenantPair(lru.Pair[TenantToken, TenantContext]):
    def __repr__(self):
        return f"{self.key}:{self.result}"

    def discard(self):
        self.result.discard()


class TenantEnvironments(lru.LRU[TenantToken, TenantContext]):
    pair = TenantPair

    def apply(self, key: TenantToken) -> TenantContext:

        return key.build_context()

    def close(self):
        with self.lock:
            if not self.cache_map.__len__():
                return
            context_item: TenantPair
            for context_item in self:
                if not context_item:
                    continue
                context_item.discard()





import unittest
import random
import threading
from concurrent.futures import thread
import peewee as og_peewee

from tenantpy.backend import (
    cache,
    mysql,
    context,
)
from tenantpy.contrib import flask


class User(flask.BaseModel):
    englishName = og_peewee.CharField()


config_1 = dict(
    host="0.0.0.0",
    port=3306,
    password="foo1",
    user="foo1",
    database="foo1",
)

config_2 = dict(
    host="0.0.0.0",
    port=3306,
    password="bar1",
    user="bar1",
    database="bar1",
)


config_3 = dict(
    url="redis://0.0.0.0:6380/0",
)


config_4 = dict(
    url="redis://0.0.0.0:6380/1",
)


class TestTenantDB(unittest.TestCase):

    def test_single(self):
        request()

    def test_mutil_thread_connect(self):
        with thread.ThreadPoolExecutor(max_workers=100) as t:
            for _ in range(1000):
                t.submit(request)

    def test_thread_local(self):
        with thread.ThreadPoolExecutor(max_workers=2) as t:
            for _ in range(100):
                t.submit(increase)


class SafeMutex(threading.local):
    def __init__(self):
        self.__dict__.update({
            "number": 0
        })


safe_mutex = SafeMutex()


def increase():
    safe_mutex.number += 1


def request():
    db_config = random.choice([config_1, config_2])
    cache_config = random.choice([config_3, config_4])
    token = context.TenantToken()
    token.organize(mysql.DataBaseMetaDataBuilder(**db_config))
    token.organize(cache.CacheMetaDataBuilder(**cache_config))
    with flask.build_context(token):
        ins = User.get()
        try:
            if db_config["database"] == "gllueweb_ats":
                assert ins.englishName == "Admin"
            elif db_config["database"] == "test_gllueweb_ats":
                assert ins.englishName == "123"
        except AssertionError:
            print(f"assert error {ins.englishName}")



import logging


class BaseLogger(logging.Logger):
    namespace = "tenant"

    def __init__(self, name, level=logging.NOTSET):
        base_class = self.__class__
        names = []
        for base in reversed(base_class.__mro__):
            namespace = getattr(base, "namespace", None)
            if namespace:
                names.append(namespace)
        names.append(name)
        super().__init__(".".join(names), level=level)



import redis

from ..backend import context


class CacheMetaData(redis.Redis):

    def discard(self):
        self.connection_pool.disconnect()


class CacheMetaDataBuilder(context.MetaDataBuilder[CacheMetaData]):

    __slots__ = ("config",)

    def __init__(self, **config):
        self.config = config

    def hash(self) -> int:
        return self.config.__str__().__hash__()

    def build(self) -> CacheMetaData:
        if "url" in self.config:
            return CacheMetaData.from_url(
                url=self.config["url"]
            )
        return CacheMetaData(**self.config)



import pymysql

try:
    from dbutils import (
        steady_db,
        pooled_db,
    )
except ImportError:
    from DBUtils import (
        SteadyDB as steady_db,
        PooledDB as pooled_db,
    )


from ..backend import context


class DataBaseMetaData(pooled_db.PooledDB):

    def steady_connection(self):
        return CustomSteadyDBConnect(
            self._creator, self._maxusage, self._setsession,
            self._failures, self._ping, True, *self._args, **self._kwargs)

    def discard(self):
        self.close()


class CustomSteadyDBConnect(steady_db.SteadyDBConnection):

    @property
    def server_version(self):
        return self._con.server_version


class DataBaseMetaDataBuilder(context.MetaDataBuilder[DataBaseMetaData]):

    __slots__ = ("config",)

    def __init__(self, **config):
        self.config = config

    def hash(self) -> int:
        return self.config.__str__().__hash__()

    def build(self) -> DataBaseMetaData:
        return DataBaseMetaData(
            creator=pymysql,
            **self.config,
        )



import random
from tenantpy.utils import lru
import unittest
from pympler import asizeof


class MyToke(lru.Token):
    def __init__(self, id_: int):
        self.id_ = id_

    def hash(self):
        return hash(self.id_)

    def __repr__(self):
        return str(self.id_)


class MyElement(lru.Element):
    def __init__(self, id_: int):
        self.switch = True
        self.id_ = id_

    def discard(self):
        self.switch = False

    def __repr__(self):
        return str(self.id_)


class MyPair(lru.Pair[MyToke, MyElement]):
    def __repr__(self):
        return f"{self.key}:{self.result}"


class MyLRU(lru.LRU[MyToke, MyElement]):
    pair = MyPair

    def apply(self, key: MyToke) -> MyElement:
        return MyElement(id_=key.id_)


class TestLRU(unittest.TestCase):

    def test_1_add(self):
        lru_ins = MyLRU(max_size=10)
        for i in range(1, 5):
            lru_ins[MyToke(i)] = MyElement(i)
            asizeof.asizeof(lru_ins)
        for _ in range(1000):
            token_num = random.choice([_ for _ in range(1, 5)])
            lru_ins[MyToke(token_num)]
            print(asizeof.asizeof(lru_ins))



import os
import gc
import signal
import enum
import tracemalloc
import linecache

__all__ = ["register_snapshot_signal"]

__old_snapshot = None


@enum.unique
class MallocEnvsKey(str, enum.Enum):
    python_trace_malloc = "PYTHONTRACEMALLOC"
    statics_key = "MALLOCSTATISTICKEY"
    statics_rank_number = "MALLOCSTATISTICRANKNUMBER"


@enum.unique
class MallocStatisticsKey(str, enum.Enum):
    traceback = 'traceback'
    filename = 'filename'
    lineno = 'lineno'


def __get_configs(key: MallocEnvsKey):
    return os.environ.get(key.value)


def take_snapshot(sig, frame=None):
    global __old_snapshot
    gc.collect()
    if not __old_snapshot:
        print(f"Start SnapShot Tracking")
        tracemalloc.start()
    snapshot = tracemalloc.take_snapshot()
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
        tracemalloc.Filter(False, "*linecache.py"),
    ))
    key = __get_configs(MallocEnvsKey.statics_key)
    rank_num = __get_configs(MallocEnvsKey.statics_rank_number)
    top_stats = snapshot.statistics(key)
    __old_snapshot = snapshot
    if not top_stats:
        return
    print(f"Top {rank_num}:")
    for index, stat in enumerate(top_stats[:int(rank_num)], 1):
        frame = stat.traceback[0]
        print("#%s: %s:%s: %.1f KiB"
              % (index, frame.filename, frame.lineno, stat.size / 1024))
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print('    %s' % line)


def register_snapshot_signal():
    signal.signal(signal.SIGTTOU, take_snapshot)



import abc
from typing import Type, Dict, Generic, TypeVar
from collections.abc import MutableMapping
import threading


class Token(abc.ABC):

    def __hash__(self):
        return self.hash()

    def __eq__(self, other: "Token"):
        return self.hash() == other.hash()

    @abc.abstractmethod
    def hash(self):
        raise NotImplementedError


class Element(abc.ABC):

    def __del__(self):
        self.discard()

    @abc.abstractmethod
    def discard(self):
        raise NotImplementedError


token = TypeVar("token", bound=Token)
element = TypeVar("element", bound=Element)


class Pair(Generic[token, element]):
    __slots__ = (
        "prev", "next", "key", "result"
    )

    def __init__(
        self,
        prev: 'Pair',
        next_v: 'Pair',
        key: token,
        result: element,
    ):
        self.prev: 'Pair' = prev
        self.next: 'Pair' = next_v
        self.key: token = key
        self.result: element = result

    def discard(self):
        self.result.discard()


class LRU(MutableMapping, Generic[token, element]):
    __slots__ = ("lock", "root", "max_size")
    cache_map: Dict[token, Pair] = {}
    pair: Type[Pair] = Pair

    @abc.abstractmethod
    def apply(self, key: token) -> element:
        raise NotImplementedError

    def __init__(
        self,
        max_size: int,
    ):
        self.lock = threading.RLock()
        self.max_size = max_size
        self.root: Pair = None

    def set_root(self, root_pair: Pair):
        if root_pair.key == self.root.key:
            return
        if root_pair.key == self.root.prev.key:
            self.root = root_pair
            return
        last = self.root.prev
        old_prev = root_pair.prev
        old_next = root_pair.next
        old_prev.next = old_next
        old_next.prev = old_prev
        self.root.prev = last.next = root_pair
        root_pair.next = self.root
        root_pair.prev = last
        self.root = root_pair

    def __getitem__(self, key: token) -> element:
        with self.lock:
            try:
                old_pair = self.cache_map[key]
                self.set_root(old_pair)
                return old_pair.result
            except KeyError:
                result = self.apply(key)
                self[key] = result
                return result

    def __setitem__(self, key: token, result: element) -> None:
        with self.lock:
            if len(self.cache_map) < self.max_size:
                new_pair = self.pair(
                    prev=None,
                    next_v=None,
                    key=key,
                    result=result,
                )
                if not self.root:
                    self.root = new_pair
                    self.root.prev = self.root
                    self.root.next = self.root
                else:
                    last = self.root.prev
                    new_pair.prev = last
                    new_pair.next = self.root
                    last.next = self.root.prev = new_pair
                    self.root = new_pair
            else:
                last = self.root.prev
                self.root = last
                del self.cache_map[last.key]
                last.key = key
                last.result = result
                new_pair = last
            self.cache_map[key] = new_pair

    def __delitem__(self, result: element) -> None:
        ...

    def __len__(self):
        return len(self.cache_map)

    def __iter__(self) -> Pair:
        yield self.root
        current_pair = self.root.next
        while current_pair.key != self.root.key:
            yield current_pair
            current_pair = current_pair.next

    def __repr__(self):
        tmp = ""
        for result in self:
            tmp += ">" + str(result)
        return tmp



from setuptools import setup, find_packages


def requirements():
    reqs = []
    with open("requirements.txt") as f:
        for line in f.readlines():
            reqs.append(line)
    return reqs


def description():
    with open("README.md") as f:
        return f.read()

setup(
    name="tenantpy",
    version="1.0.5",
    author="zhouhao",
    python_requires='>=3.7.0',
    author_email="zhouhao19931002@hotmail.com",
    long_description=description(),
    long_description_content_type="text/markdown",
    maintainer="zhouhao",
    description="Tenant Management For Web Application",
    packages=find_packages(exclude=("tests",)),
    install_requires=requirements(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.7"
    ]
)
